﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module common.cacheheadertable;

// キャッシュインデックス SQLiteバージョン

private import std.string;
private import std.utf;
private import std.stdio;
private import std.random;

private import pes.prompt;
private import pes.wstring;
private import pes.random;
private import pes.exception;
private import pes.net.winny.key;
private import pes.system.locker;
private import pes.system.delay;
private import pes.system.semaphore;
private import pes.data.sqlite3;

private import common.configuration;
private import worker.cachebitmap;


private const char[] CACHE_HEADER_TABLE_CREATE = 
	"CREATE TABLE cache_header_table(hash blob, size int64, name text, mtime int, referred_count int, bitmap blob, completed int)";

private const char[] CACHE_HEADER_INDEX_CREATE = 
	"CREATE UNIQUE INDEX cache_header_index on cache_header_table(hash, size)";

private const char[] CACHE_HEADER_COMPLETED_SELECT = 
	"SELECT * FROM cache_header_table WHERE completed=1";

private const char[] CACHE_HEADER_COUNT = 
	"SELECT count(*) FROM cache_header_table";

private const char[] CACHE_HEADER_EXISTS = 
	"SELECT count(*) FROM cache_header_table WHERE hash=? AND size=?";
	
private const char[] CACHE_HEADER_DELETE = 
	"DELETE FROM cache_header_table WHERE hash=? AND size=?";

private const char[] CACHE_HEADER_SELECT = 
	"SELECT * FROM cache_header_table WHERE hash=? AND size=?";

private const char[] CACHE_HEADER_INSERT = 
	"INSERT INTO cache_header_table(hash, size, name, mtime, referred_count, bitmap, completed) values(?,?,?,?,?,?,?)";

private const char[] CACHE_HEADER_UPDATE = 
	"UPDATE cache_header_table SET name = ?, mtime = ?, referred_count = ?, bitmap = ?, completed = ? WHERE hash = ? and size = ?";

private const char[] CACHE_HEADER_REF_ADD =
	"UPDATE cache_header_table SET referred_count = referred_count + ? WHERE hash=? AND size=?";

private const char[] CACHE_HEADER_COMPLEITE_SELECT = 
	"SELECT * FROM cache_header_table WHERE completed=1";

private const char[] CACHE_HEADER_SEARCH_NAME =
"SELECT * FROM cache_header_table WHERE name like('%'|| ? ||'%') LIMIT ?";

private const char[] CACHE_HEADER_SEARCH_ALL =
"SELECT * FROM cache_header_table LIMIT ?";


private class SearchQuery
{
protected:
	char[] 	m_query;
	bool	m_first;
	
public:
	this()
	{
		m_query = "SELECT * FROM cache_header_table WHERE ";
		m_first = true;
	}
	
	void and(char[] keyword)
	{
		char[] add;
		
		if (! m_first) {
			add = " AND ";
		} else {
			m_first = false;
		}
		
		m_query ~= add ~ std.string.format("name like ('%%%s%%')",
												escapeSQLText(keyword));
	}
	
	void not(char[] keyword)
	{
		char[] add;
		
		if (! m_first) {
			add = " AND ";
		} else {
			m_first = false;
		}
		m_query ~= add ~ std.string.format("name not like('%%%s%%')",
												escapeSQLText(keyword));
	}
	
	char[] text()
	{
		return m_query ~ " LIMIT ?";
	}
	
	void reset()
	{
		m_query = null;
	}
}

private char[] makeQuery(char[] input)
{
	char[][] scrap = std.string.split(input);
	SearchQuery query = new SearchQuery;
	
	foreach (char[] part; scrap) {
		if (part.length == 0 ) {
			continue;
		}
		if (part[0] == '-') {
			query.not(part[1 .. length]);
		} else {
			query.and(part);
		}
	}
	
	return query.text;
}



// キャッシュヘッダ

class CacheHeader {
	ubyte[]		hash;
	ulong		dataSize;
	char[]		fileName;
	uint		modifiedTime;
	uint		referredCount;
	CacheBitmap	bitmap;
	byte		completed;
	
	this()
	{
		bitmap.clear();
	}
	
	bool isComplete()
	{
		return completed == 1;
	}
	
	char[] toString()
	{
		return std.string.format("%s,%s,%s,%s,%s", hash, dataSize, fileName, modifiedTime, referredCount);
	}
}

// キャッシュヘッダテーブル
class CacheHeaderTable
{
protected:
	SQLite				m_db;
	char[]				m_file;
	SemaphoreResource 	m_semaphore;
	semaphore			m_id;
	
	CacheBitmap toBitmap(ubyte[] bitmap)
	{
		if (bitmap.length != CacheBitmap.sizeof) {
			CacheBitmap zeromap;
			return zeromap;
		}
		
		return *cast(CacheBitmap *)bitmap.ptr;
	}
	
public:
	void signal()
	{
		m_semaphore.signal(m_id);
	}
	
	void release()
	{
		m_semaphore.release(m_id);
	}
	
	
	this(char[] path, SemaphoreResource resource, semaphore id)
	{
		m_semaphore	= resource;
		m_id		= id;
		m_file 		= path;
		
		signal();
		try {
			m_db = new SQLite(m_file);
			m_db.exec(CACHE_HEADER_TABLE_CREATE);
			m_db.exec(CACHE_HEADER_INDEX_CREATE);
		} catch (SQLiteException) {
			// ignore
		} finally {
			release();
		}
	}
	
	void close()
	{
		signal();
		try {
			m_db.close();
		} finally {
			release();
		}
	}
	
	uint length()
	{
		uint count;
		
		signal();
		try {
			auto SQLiteStatement stmt = m_db.prepare(CACHE_HEADER_COUNT);
			stmt.step();
			if (stmt.hasRow) {
				count = stmt.fetch!(int)(0);
			}
//			stmt.finalize();
		} finally {
			release();
		}
		
		return count;
	}
	
	bool exists(ubyte[] hash, ulong data_size)
	{
		uint count;
		
		signal();
		try {
			auto SQLiteStatement stmt = m_db.prepare(CACHE_HEADER_EXISTS);
			stmt.bind(1, hash, data_size);
			stmt.step();
			if (stmt.hasRow) {
				count = stmt.fetch!(int)(0);
			}
//			stmt.finalize();
		} finally {
			release();
		}
		
		return count != 0;
	}
	
	void set(CacheHeader header)
	{
		signal();
		try {
			m_db.begin();
			
			if (header.fileName.length == 0
				&& exists(header.hash, header.dataSize))
			{
				CacheHeader old = get(header.hash, header.dataSize);
				header.fileName = old.fileName;
			}
			auto SQLiteStatement stmt = m_db.prepare(CACHE_HEADER_INSERT);
			
			stmt.bind(1, header.hash, header.dataSize, header.fileName, header.modifiedTime,
									header.referredCount, header.bitmap.toBlob, header.completed);
			stmt.step();
		} finally {
//			stmt.finalize();
			m_db.commit();
			release();
		}
	}
	
	void update(CacheHeader header)
	{
		signal();
		try {
			m_db.begin();
			CacheHeader old = get(header.hash, header.dataSize);
			if (header.fileName.length == 0) {
				CacheHeader old = get(header.hash, header.dataSize);
				header.fileName = old.fileName;
			}
			header.bitmap.merge(old.bitmap);
			
			auto SQLiteStatement	stmt = m_db.prepare(CACHE_HEADER_UPDATE);
			
			stmt.bind(1, header.fileName, header.modifiedTime,
							header.referredCount, header.bitmap.toBlob, header.completed,
							header.hash, header.dataSize);
			stmt.step();
		} finally {
//			stmt.finalize();
			m_db.commit();
			release();
		}
	}
	
	CacheHeader get(ubyte[] hash, ulong data_size)
	{
		uint	pos = 0;
		CacheHeader header = new CacheHeader;
		
		signal();
		try {
			auto SQLiteStatement stmt = m_db.prepare(CACHE_HEADER_SELECT);
			stmt.bind(1, hash, data_size);
			stmt.step();
			if (!stmt.hasRow) {
				throw new NotFound("cache header");
			}
			header.hash 			= stmt.fetch!(ubyte[])(pos++).dup;
			header.dataSize 		= stmt.fetch!(ulong)(pos++);
			header.fileName 		= stmt.fetch!(char[])(pos++).dup;
			header.modifiedTime 	= stmt.fetch!(uint)(pos++);
			header.referredCount	= stmt.fetch!(uint)(pos++);
			header.bitmap 			= toBitmap(stmt.fetch!(ubyte[])(pos++));
			header.completed		= stmt.fetch!(ubyte)(pos++);
			
//			stmt.finalize();
		} finally {
			release();
		}
		
		return header;
	}
	
	void referred(ubyte[] hash, ulong data_size, uint block_count)
	{
		signal();
		try {
			auto SQLiteStatement stmt = m_db.prepare(CACHE_HEADER_REF_ADD);
			stmt.bind(1, block_count, hash, data_size);
			
			stmt.step();
//			stmt.finalize();
		} finally {
			release();
		}
	}
	
	NyKeyInformation[] completeCacheKeys(Configuration config)
	{
		uint	pos = 0;
		NyKeyInformation[] cache;
		NyKeyInformation key;
		
		signal();
		try {
			auto SQLiteStatement stmt = m_db.prepare(CACHE_HEADER_COMPLEITE_SELECT);
			
			for (stmt.step(); stmt.hasRow; stmt.step()) {
				pos = 0;
				key = new NyKeyInformation;
				key.keyAddress			= config.address;
				key.keyPort				= config.fileSharingPort;
				key.hash				= stmt.fetch!(ubyte[])(pos++).dup;
				key.fileSize			= stmt.fetch!(ulong)(pos++);
				key.fileName			= stmt.fetch!(char[])(pos++).dup;
				key.modifiedTime		= stmt.fetch!(uint)(pos++);
				key.referredBlock 		= stmt.fetch!(uint)(pos++);
				key.trip				= config.trip;
				key.fileNameMBSLength 	= utf8ToMBS(key.fileName).length;
				key.keyVersion			= 5;
				key.timer				= 1500 + (rand() % 300);
				
				cache ~= key;
			}
		} finally {
//			stmt.finalize();
			release();
		}
		
		return cache;
	}
	
	void remove(ubyte[] hash, ulong data_size)
	{
		signal();
		try {
			auto SQLiteStatement stmt = m_db.prepare(CACHE_HEADER_DELETE);
			stmt.bind(1, hash, data_size);
			stmt.step();
//			stmt.finalize();
		} finally {
			release();
		}
	}
	
	CacheHeader[] searchName(char[] keyword, uint max = uint.max)
	{
		CacheHeader[] result;
		uint i, pos;
		
		signal();
		try {
			SQLiteStatement stmt;
			
			if (keyword == "*") {
				stmt = m_db.prepare(CACHE_HEADER_SEARCH_ALL);
				stmt.bind(1, max);
			} else {
				stmt = m_db.prepare(makeQuery(keyword));
				stmt.bind(1, max);
			}
			scope(exit) stmt.finalize();
//			m_db.begin();
			
			for (stmt.step(); stmt.hasRow; stmt.step()) {
				CacheHeader header 	= new CacheHeader;
				pos = 0;
				header.hash 			= stmt.fetch!(ubyte[])(pos++).dup;
				header.dataSize 		= stmt.fetch!(ulong)(pos++);
				header.fileName 		= stmt.fetch!(char[])(pos++).dup;
				header.modifiedTime 	= stmt.fetch!(uint)(pos++);
				header.referredCount	= stmt.fetch!(uint)(pos++);
				header.bitmap 			= toBitmap(stmt.fetch!(ubyte[])(pos++));
				header.completed		= stmt.fetch!(ubyte)(pos++);
				
				result ~= header;
				
				if (++i >= max) {
					break;
				}
			}
//			stmt.finalize();
//			m_db.end();
		} finally {
			release();
		}
		
		return result;
	}
}

