﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module worker.downloader;

// ダウンローダー

private import std.socket;
private import std.stdio;
private import std.thread;
private import std.random;
private import std.file;
private import std.gc;
private import pes.net.winny.all;
private import pes.conv;
private import pes.rc4;
private import pes.checksum;
private import pes.data.sqlite3;
private import pes.system.all;
private import common.task;
private import common.environment;
private import common.cacheresolve;
private import common.workertable;
private import common.packet;

debug(download)
{
	import pes.prompt;
}

private const char[] CACHE_BLOCK_TABLE_CREATE = 
	"CREATE TABLE cache_block_table(no int64, hash blob, block blob)";

private const char[] CACHE_BLOCK_INDEX_CREATE =
	"CREATE UNIQUE INDEX cache_block_index on cache_block_table(no)";

private const char[] CACHE_BLOCK_INSERT =
	"INSERT INTO cache_block_table(no, hash, block) values(?, ?, ?)";
	
private const char[] CACHE_BLOCK_TEMP_CREATE = 
	"CREATE TEMP TABLE cache_block_temp(no int64, hash blob, block blob)";

private const char[] CACHE_BLOCK_TEMP_INSERT =
	"INSERT INTO cache_block_temp(no, hash, block) values(?, ?, ?)";
	
private const char[] CACHE_BLOCK_TEMP_COPY =
	"INSERT INTO cache_block_table (no, hash, block) "
	"select * from cache_block_temp where no not in(select no from cache_block_table)";

class DownloadFile {
	ubyte[]	hash;
	uint	fileSize;
	uint	beginBlock;
	uint	count;
	
	this(ubyte[] hash_md5, uint file_size, uint begin_block, uint count)
	{
		hash 		= hash_md5.dup;
		fileSize	= file_size;
		beginBlock	= begin_block;
		this.count	= count;
	}
}

class Downloader: WorkerTask
{
public:
	
protected:
	enum DownloadState: uint
	{
		OK				= 0,		// ok
		CONNECT_ERROR,				// connect error
		TIMEOUT,					// response timeout
		BUSY,						// busy node
		NG,							// nazo
	}
	
	InternetAddress m_node;
	DownloadState	m_state;
	Environment		s_environ;
	DownloadFile	m_request;
	CacheHeader		m_header;
	DownloadItem	m_item;
	uint			m_download_blocks;
	Connection		m_connection;
	
	int mainLoop()
	{
		CacheHeader header = null;
		
		gcLimit = 16;
		
		try {
			char[] path = cacheResolve(m_request.hash, s_environ.configuration);
			if (!std.file.exists(path)) {
				header = makeNewCache(path, m_request);
			} else {
				header = s_environ.cacheHeader.get(m_request.hash, m_request.fileSize);
			}
			m_header = header;
			
			name = header.fileName;
			
			download(header, path);
		} catch (Object e) {
			m_state = DownloadState.NG;
			debug(download)
				mbswritef(derr, "Downloader: Exception %s\n", e.toString);
		}
		
		std.gc.fullCollect();
		
		return 0;
	}
	
	CacheHeader makeNewCache(char[] path, DownloadFile item)
	{
		scope(exit) db.close();
		CacheHeader header	= new CacheHeader;
		SQLite db			= new SQLite(path);
		NyKeyInformation[]	key;
		
		db.exec(CACHE_BLOCK_TABLE_CREATE);
		db.exec(CACHE_BLOCK_INDEX_CREATE);
		db.close();
		
		if (s_environ.cacheHeader.exists(item.hash, item.fileSize)) {
			s_environ.cacheHeader.remove(item.hash, item.fileSize);
		}
		
		key = s_environ.keyTable.searchHash(item.hash);
		if (key.length == 0) {
			header.fileName	= m_item.name;
			header.hash 	= m_item.hash;
			header.dataSize	= m_item.dataSize;
			debug(download)
					mbswritef(derr, "Downloader: new Hash by item %s,%s\n", item.hash, item.fileSize);
		} else {
			header.hash				= key[0].hash;
			header.dataSize			= key[0].fileSize;
			header.fileName			= key[0].fileName;
			header.modifiedTime		= key[0].modifiedTime;
			header.referredCount	= key[0].referredBlock;
			debug(download)
					mbswritef(derr, "Downloader: new Hash by key %s,%s\n", key[0].hash, key[0].fileSize);
		}
		header.bitmap.clear();
		s_environ.cacheHeader.set(header);
		
		return header;
	}
	
	void download(CacheHeader cache_header, char[] cache)
	{
		NyProtocolHeader		header;
		Connection				conn 		= new Connection;
		NyFileRequest 			request		= new NyFileRequest;
		NyFileResponse			response	= new NyFileResponse;
		SQLite					db			= new SQLite(cache);
		SQLiteStatement			stmt;
		NyRawCommand 			command;
		ubyte[]					key;
		ubyte[]					block		= new ubyte[0x10000];
		uint					retry_timeout	= 0;
		uint					blockno 		= 0;
		
		workMax(m_request.count * 0x10000);
		try {
			db.exec(CACHE_BLOCK_TEMP_CREATE); // テンポラリ作成
			
			debug(download) mbswritef(derr, "Downloader: コネクト開始\n");
			try {
				NodeDetails node = new NodeDetails(m_node);
				header = conn.connect(s_environ.configuration,
										node, NyConnectionType.LinkType.TRANSFER, 
										s_environ.configuration.downloadTimeout);
			} catch (ConnectionTimeOut) {
				m_state = DownloadState.TIMEOUT;
				return;
			} catch (SocketException) {
				m_state = DownloadState.CONNECT_ERROR;
				return;
			}
			m_connection = conn;
			
			conn.lock();
			s_environ.dlDataConnection.set(conn.aaKey, conn);
//			if (!(header.major == NyProtocolHeader.WINNY_MAJOR 
//				&& header.minor == NyProtocolHeader.WINNY_MINOR))
			if (header.minor != NyProtocolHeader.WINNY_MINOR) {
				m_state = DownloadState.CONNECT_ERROR;
				conn.connection.sendCommand(new NyLowVersion);
				return;
			}
			debug(download) mbswritef(derr, "Downloader: コネクト完了\n");
			
			request.taskId 			= rand() % 100000;
			request.beginBlock		= m_request.beginBlock;
			request.requestCount	= m_request.count;
			request.hash			= m_request.hash;
			request.fileSize		= m_request.fileSize;
			
			debug(download) 
			mbswritef(derr, "Downloader: ダウンロード開始 %s (block:%s-, count:%s)\n", 
				 binaryToHexString(request.hash), request.beginBlock, request.requestCount);
			
//			db.begin();
			stmt = db.prepare(CACHE_BLOCK_TEMP_INSERT);
			
			delayTask(100);
			debug(download) mbswritef(derr, "Downloader: リクエスト\n");
			conn.connection.sendCommand(request);
			
			uint i = 0;
			for (blockno = request.beginBlock; blockno < request.beginBlock + request.requestCount; /* none */) {
				debug(download) mbswritef(derr, "Downloader: 受信待ち: %s\n", blockno);
				try {
					pollTerminateEvent(); // 中断チェック
					
					conn.connection.timeout(s_environ.configuration.downloadTimeout, 0);
					command = conn.connection.receiveCommand();
					switch (command.commandCode) {
					case NyCommandCode.FILE_RESPONSE:
						// ok
						break;
						
					case NyCommandCode.CONNECTED_LIMITATION:
						// ビジー
						m_state = DownloadState.BUSY;
						return;
						
					case NyCommandCode.QUERY:
						sendMail(MailId.CONTROL_REQUEST,
							Mail.makeRequest(MessageId.ROUTING,
								new ReceivedCommand(command, new NodeDetails(m_node))));
						continue;
						
					case NyCommandCode.REJECT:
					case NyCommandCode.LIAR:
					case NyCommandCode.SLOW_RATE:
					case NyCommandCode.LOW_VERSION:
					case NyCommandCode.WRONG_LISTENING_PORT:
						// 拒否系
						m_state = DownloadState.CONNECT_ERROR;
						return;
						
					case NyCommandCode.CLOSE:
						m_state = DownloadState.NG;
						return;
					default:
						debug(download)
							writef("command Download: %s\n", command.toString);
						continue;
					}
				} catch (ConnectionTimeOut) {
					if (++retry_timeout > 3) {
						m_state = DownloadState.TIMEOUT;
						return;
					}
					continue;
				}
				retry_timeout = 0;
				
				pollTerminateEvent(); // 中断チェック
				
				debug(download) {
					mbswritef(derr, "Downloader: ダウンロード %s\n", blockno);
					mbswritef(derr, "Downloader: %s\n", command.toString);
				}
				if (blockno == 0) {
					key = NyFileResponse.HEADER_KEY;
				} else {
					key = cast(ubyte[])std.string.format("%d %s", blockno, binaryToHexString(request.hash));
				}
				
				response.unpack(command);
				command.clearBuffer();
				
				if (response.taskId != request.taskId) {
					log(std.string.format("%s(%s) task ID error", binaryToHexString(request.hash), blockno));
					debug(download) 
						mbswritef(derr, "Downloader: タスクIDエラー\n");
//					throw new Exception("Downloader: タスクID不一致");
					m_state = DownloadState.CONNECT_ERROR;
					return;
				}
				block[0 .. response.dataSize] = response.fileData;
				pes.rc4.RC4.crypt(key, block[0 .. response.dataSize]);
				
				if (!checkMD5Sum(response.hash, block[0 .. response.dataSize])) {
					debug(download) 
						mbswritef(derr, "Downloader: サムエラー\n");
					log(std.string.format("%s(%s) md5 sum error", binaryToHexString(request.hash), blockno));
//					throw new Exception("Downloader: ハッシュ不一致");
					m_state = DownloadState.CONNECT_ERROR;
					return;
				}
				
				debug(download) mbswritef(derr, "Downloader: ブロック受信(%s)\n", blockno);
				if (!cache_header.bitmap.isSet(blockno)) {
					stmt.bind(1, blockno, response.hash, response.fileData);
					stmt.step();
					stmt.reset();
					cache_header.bitmap.set(blockno);
					m_item.setBit(blockno);
				} else {
					debug(download) 
						mbswritef(derr, "Downloader: キャッシュなぜか存在!!\n");
					log(std.string.format("%s(%s) cache block no already exists",
						binaryToHexString(request.hash), blockno));
					throw new Exception("Downloader: キャッシュなぜか存在");// おかしい
				}
				debug(download) mbswritef(derr, "Downloader: 保存完了(%s)\n", blockno);
				
				// item incl
				
				++m_download_blocks;
				++blockno;
				work(0x10000);
				
				this.fullCollect();
				delayTask(10);
			}
			debug(download) 
				mbswritef(derr, "Downloader: ダウンロード 完了\n");
			
			m_state = DownloadState.OK;
		
		} catch (Exception e) {
			debug(download)
				mbswritef(derr, "Downloader: Exception %s\n", e.toString);
			m_state = DownloadState.NG;
			
			return;
		} finally {
			delete block;
			delete response;
			block = null;
			response = null;
			if (conn !is null) {
				if (conn.nodeDetails !is null
					&& s_environ.dlDataConnection.exists(conn.aaKey))
				{
					s_environ.dlDataConnection.remove(conn.aaKey);
				}
			}
			
			if (stmt !is null) {
				stmt.finalize();
			}
			synchronized(m_item.cacheLock) {
				db.begin();
				db.exec(CACHE_BLOCK_TEMP_COPY);
				db.commit();
				db.close();
			}
			
			// キャッシュインデックス更新
			updateCacheIndex(cache_header, blockno - request.beginBlock);
		}
	}
	
	void updateCacheIndex(CacheHeader cache_header, uint inc_blocks)
	{
		s_environ.cacheHeader.signal();
		try {
			CacheHeader	old 	= s_environ.cacheHeader.get(cache_header.hash, cache_header.dataSize);
			
			// マージしとく
			cache_header.referredCount = old.referredCount + inc_blocks;
			
			s_environ.cacheHeader.update(cache_header);
			//delete old;
			old = null;
		} finally {
			s_environ.cacheHeader.release();
		}
	}
public:
	this(TaskResource resource, InternetAddress node, DownloadFile request,
					DownloadItem item, Environment environ)
	{
		super(ANONYMOUS_MAIL, resource);
		m_node 		= node;
		m_request	= request;
		s_environ	= environ;
		m_item		= item;
		m_state		= DownloadState.OK;
	}
	
	DownloadFile requestFile()
	{
		return m_request;
	}
	
	ubyte[] hash()
	{
		return m_request.hash;
	}
	
	CacheHeader cacheHeader()
	{
		return m_header;
	}
	
	InternetAddress address()
	{
		return m_node;
	}
	
	DownloadState exitState()
	{
		return m_state;
	}
	
	override uint rate()
	{
		if (m_connection !is null
			&& m_connection.connection !is null) 
		{
			return m_connection.connection.rate();
		}
		
		return super.rate();
	}
	
	uint downloadBlocks()
	{
		return m_download_blocks;
	}
}

